/*
 * Z80SIM  -  a Z80-CPU simulator
 *
 * Common I/O devices used by various simulated machines
 *
 * Copyright (C) 2014-2017 by Udo Munk
 * Copyright (C) 2017 changes by Greg Sydney-Smith
 *
 * Emulation of a Cromemco 4FDC/16FDC S100 board
 *
 * History:
 * 20-DEC-2014 first version
 * 28-DEC-2014 second version with 16FDC, CP/M 2.2 boots
 * 01-JAN-2015 fixed 16FDC, machine now also boots CDOS 2.58 from 8" and 5.25"
 * 01-JAN-2015 fixed frontpanel switch settings, added boot flag to fp switch
 * 12-JAN-2015 implemented dummy write track, so that programs won't hang
 * 22-JAN-2015 fixed buggy ID field, fake index pulses for 300/360 RPM
 * 26-JAN-2015 implemented write track to format SS/SD disks properly
 * 02-FEB-2015 implemented DS/DD disk formats
 * 05-FEB-2015 implemented DS/SD disk formats
 * 06-FEB-2015 implemented write track for all formats
 * 12-FEB-2015 implemented motor control, so that a 16FDC is recogniced by CDOS
 * 20-FEB-2015 bug fixes for 1.25 release
 * 08-MAR-2016 support user path for disk images
 * 13-MAY-2016 find disk images at -d <path>, ./disks and DISKDIR
 * 22-JUL-2016 added support for read only disks
 * 22-JUN-2017 added reset function
 * 26-JUL-2017 fixed buggy index pulse implementation
 * 14-NOV-2017 gss. Mods for Win32.
 * 26-NOV-2017 gss. Don't refuse Cmds whilst BUSY. And dbg changes, fdc_tick()

 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "sim.h"
#include "simglb.h"
#include "cromemco-fdc.h"

#ifdef _WIN32
#define OPENRO O_RDONLY|O_BINARY
#define OPENRW O_RDWR|O_BINARY
#else
#define OPENRO O_RDONLY
#define OPENRW O_RDWR
#endif

#define DISKS 4

/* internal state of the fdc */
#define FDC_IDLE	0	/* idle state */
#define FDC_READ	1	/* reading sector */
#define FDC_WRITE	2	/* writing sector */
#define FDC_READADR	3	/* read address */
#define FDC_WRTTRK	4	/* write (format) track */

#define SEC_SZSD	128	/* sector size SD */
#define SEC_SZDD	512	/* sector size DD */
#define TRK8		77	/* # of tracks 8" */
#define SPT8SD		26	/* # of sectors per track 8" SD */
#define SPT8DD		16	/* # of sectors per track 8" DD */
#define TRK5		40	/* # of tracks 5.25" */
#define SPT5SD		18	/* # of sectors per track 5.25" SD */
#define SPT5DD		10	/* # of sectors per track 5.25" DD */

/*
 * disk definitions 5.25"/8" drives, single/double density,
 * single/double sided
 */
enum Disk_type    { SMALL, LARGE, UNKNOWN };
enum Disk_density { SINGLE, DOUBLE };
enum Disk_sides   { ONE, TWO };
enum Disk_mode    { READWRITE, READONLY };

typedef struct {
	char *fn;			/* filename of disk image */
	enum Disk_type disk_t;		/* drive type 5.25" or 8" */
	enum Disk_density disk_d;	/* disk density, single or double */
	enum Disk_sides disk_s;		/* drive sides, 1 or 2 */
	int tracks;			/* # of tracks */
	int sectors;			/* # sectors on tracks > 0 side 0 */
	int sec0;			/* # sectors on track 0, side 0 */
	enum Disk_mode disk_m;		/* R/W or R/O mode */
} Diskdef;

//     BYTE fdc_flags = 16|64;	/* FDC flag register, no autoboot */
       BYTE fdc_flags = 16;	/* FDC flag register, autoboot */
static BYTE fdc_stat;		/* FDC status register */
static BYTE fdc_aux = 3;	/* FDC auxiliary status, default: boot from A */
static BYTE fdc_track = 0;	/* FDC track register */
static BYTE fdc_sec = 1;	/* FDC sector register */
static BYTE fdc_cmd;		/* FDC command last sent */
static int step_dir = -1;	/* last stepping direction */
       enum Disk_type dtype;	/* disk type set */
static enum Disk_density ddens;	/* disk density set */
static int disk = 0;		/* current selected disk # */
static int side = 0;		/* disk side */
static int secsz = SEC_SZSD;	/* current used sector size */
static int state;		/* internal fdc state */
static int dcnt;		/* data counter read/write */
static int mflag;		/* multiple sectors flag */
static char fn[4096];		/* path/filename for disk image */
static int fd;			/* fd for disk i/o */
static BYTE buf[SEC_SZDD];	/* buffer for one sector */
//9
static int index_pulse = 0;	/* disk index pulse */
static int autowait;		/* autowait flag */
static int motoron;		/* motor on flag */
       int motortimer;		/* motor on timer */
static int headloaded;		/* head loaded flag */

/* these are our disk drives, 8" SS SD initially */
static Diskdef disks[DISKS] = {
    { "drivea.dsk", LARGE, SINGLE, ONE, TRK8, SPT8SD, SPT8SD, READWRITE },
    { "driveb.dsk", LARGE, SINGLE, ONE, TRK8, SPT8SD, SPT8SD, READWRITE },
    { "drivec.dsk", LARGE, SINGLE, ONE, TRK8, SPT8SD, SPT8SD, READWRITE },
    { "drived.dsk", LARGE, SINGLE, ONE, TRK8, SPT8SD, SPT8SD, READWRITE }
};
static int dskActv[DISKS];

/*
16fdc switches & defaults:
sw1 OFF RDOS-II is not defeated (ROM is present. not used here.)
sw2 ON  RDOS-II disabled after boot (allows "out (40),a" to disable ROM. we're RAM - not used)
sw3 ON  operating system loads automatically  (autoboot)
sw4 OFF operating system can format diskettes
sw5 OFF baud rate not pre-set (probably HW)
sw6 OFF unused
sw7 OFF \ RDOS 2.52 boot drive: 00=A, 01=B 10=C 11=D
sw8 OFF /

dbg:
 0 off
>0 commands (& their completion for those that set BUSY)
>1 OUTs to significant ports (eg track, sector; not data stream out)
>2 polling (INs of status ports, where the status changes
*/
static int dbg=0;
static int switches[1+8]= { 99,0,1,1,0,0,0,0,0 }; // 0xf9; // 64=autoboot 16

int cromemco_fdc_config(char *name, char *val) {
    int i;
    char *names[] = { "debug", "sw1", "sw2", "sw3", "sw4", "sw5", "sw6", "sw7", "sw8", NULL };

    if (val==NULL) val="0";
    for (i=0; names[i]!=NULL && stricmp(name,names[i])!=0; i++) ;
    switch (i) {
    case 0: dbg    = atoi(val); break;
    case 1:
    case 2:
    case 3:
    case 4:
    case 5:
    case 6:
    case 7:
    case 8: switches[i]= atoi(val); break;
    }
    return 0;
}

int cromemco_fdc_init(void) {
    int i;

    for (i=0; i<DISKS; i++) dskActv[i]=0;
    fdc_flags = (switches[4]?0:16) | (switches[3]?0:64) ;
    fdc_aux   = 3-((switches[7]?2:0)+(switches[8]?1:0));
    return 0;
}


/*
 * find and set path for disk images
 */
void dsk_path(void) {
    struct stat sbuf;

    /* if option -d is used disks are there */
    if (diskdir != NULL) {
        strcpy(fn, diskd);
    } else {
        /* if not first try ./disks */
        if ((stat("./disks", &sbuf) == 0) && S_ISDIR(sbuf.st_mode)) {
            strcpy(fn, "./disks");
        /* nope, then DISKSDIR as set in Makefile */
        } else {
            strcpy(fn, DISKSDIR);
        }
    }
}

/*
 * configure drive and disk geometry from image file size
 */
void config_disk(int fd)
{
    struct stat s;

    fstat(fd, &s);
    disks[disk].disk_m = (s.st_mode & S_IWUSR) ? READWRITE : READONLY;

    switch (s.st_size) {
    case 92160:		/* 5.25" SS SD */
        disks[disk].disk_t = SMALL;
        disks[disk].disk_d = SINGLE;
        disks[disk].disk_s = ONE;
        disks[disk].tracks = TRK5;
        disks[disk].sectors = SPT5SD;
        disks[disk].sec0 = SPT5SD;
        break;

    case 184320:		/* 5.25" DS SD */
        disks[disk].disk_t = SMALL;
        disks[disk].disk_d = SINGLE;
        disks[disk].disk_s = TWO;
        disks[disk].tracks = TRK5;
        disks[disk].sectors = SPT5SD;
        disks[disk].sec0 = SPT5SD;
        break;

    case 201984:		/* 5.25" SS DD */
        disks[disk].disk_t = SMALL;
        disks[disk].disk_d = DOUBLE;
        disks[disk].disk_s = ONE;
        disks[disk].tracks = TRK5;
        disks[disk].sectors = SPT5DD;
        disks[disk].sec0 = SPT5SD;
        break;

    case 406784:		/* 5.25" DS DD */
        disks[disk].disk_t = SMALL;
        disks[disk].disk_d = DOUBLE;
        disks[disk].disk_s = TWO;
        disks[disk].tracks = TRK5;
        disks[disk].sectors = SPT5DD;
        disks[disk].sec0 = SPT5SD;
        break;

    case 256256:		/* 8" SS SD */
        disks[disk].disk_t = LARGE;
        disks[disk].disk_d = SINGLE;
        disks[disk].disk_s = ONE;
        disks[disk].tracks = TRK8;
        disks[disk].sectors = SPT8SD;
        disks[disk].sec0 = SPT8SD;
        break;

    case 512512:		/* 8" DS SD */
        disks[disk].disk_t = LARGE;
        disks[disk].disk_d = SINGLE;
        disks[disk].disk_s = TWO;
        disks[disk].tracks = TRK8;
        disks[disk].sectors = SPT8SD;
        disks[disk].sec0 = SPT8SD;
        break;

    case 625920:		/* 8" SS DD */
        disks[disk].disk_t = LARGE;
        disks[disk].disk_d = DOUBLE;
        disks[disk].disk_s = ONE;
        disks[disk].tracks = TRK8;
        disks[disk].sectors = SPT8DD;
        disks[disk].sec0 = SPT8SD;
        break;

    case 1256704:		/* 8" DS DD */
        disks[disk].disk_t = LARGE;
        disks[disk].disk_d = DOUBLE;
        disks[disk].disk_s = TWO;
        disks[disk].tracks = TRK8;
        disks[disk].sectors = SPT8DD;
        disks[disk].sec0 = SPT8SD;
        break;

    default:
        disks[disk].disk_t = UNKNOWN;
        break;
    }
}

/*
 * calculate disk image seek position for track/sector/side
 */
long get_pos(void)
{
    long pos = -1L;

    /* single sided */
    if (disks[disk].disk_s == ONE) {

        /* single density */
        if (disks[disk].disk_d == SINGLE) {
          pos = (fdc_track * disks[disk].sectors + fdc_sec - 1) * SEC_SZSD;

        /* double density */
        } else {
            if (fdc_track == 0) {
                pos = (fdc_sec - 1) * SEC_SZSD;
            } else {
                pos = (disks[disk].sec0 * SEC_SZSD) +
                  ((fdc_track - 1) * disks[disk].sectors * SEC_SZDD) +
                  ((fdc_sec - 1) * SEC_SZDD);
            }
        }

    /* double sided */
    } else {

        /* single density */
        if (disks[disk].disk_d == SINGLE) {
            pos = fdc_track * 2 * disks[disk].sectors * SEC_SZSD;
            if (side == 0) {
                pos += (fdc_sec - 1) * SEC_SZSD;
            } else {
                pos += disks[disk].sectors * SEC_SZSD;
                pos += (fdc_sec - 1) * SEC_SZSD;
            }

        /* double density */
        } else {
            if ((fdc_track == 0) && (side == 0)) {
                pos = (fdc_sec - 1) * SEC_SZSD;
                goto done;
            }
            if ((fdc_track == 0) && (side == 1)) {
                pos = disks[disk].sec0 * SEC_SZSD + (fdc_sec - 1) * SEC_SZDD;
                goto done;
            }
            pos = disks[disk].sec0 * SEC_SZSD + disks[disk].sectors * SEC_SZDD;
            pos += (fdc_track - 1) * 2 * disks[disk].sectors * SEC_SZDD;
            if (side == 1)
                pos += disks[disk].sectors * SEC_SZDD;
            pos += (fdc_sec - 1) * SEC_SZDD;
        }
    }

done:
    return(pos);
}


/* (IN 34H)
 *	4FDC		16FDC
 * D7	DRQ		DRQ
 * D6	!BOOT		!BOOT
 * D5	HEADLOAD	HEADLOAD
 * D4	unassigned	!INHIBIT INIT
 * D3	unassigned	MOTOR ON
 * D2	unassigned	MOTOR TIMEOUT
 * D1	unassigned	AUTOWAIT TIMEOUT
 * D0	EOJ		EOJ
 */
BYTE cromemco_fdc_diskflags_in(void)
{
    BYTE ret;
    static int last = -1;

    /* reset EOJ after it was read */
    ret = fdc_flags;
    fdc_flags &= ~1;

    /* process autowait timeout */
    if (!(ret & 1) && !(ret & 128) && (autowait == 1)) {
        ret |= 2; // not in a 4FDC (unused)
    }

    /* motor status & timer */
    if (motoron) {
        if (motortimer > 0)
            ret |= 8; // ditto
        else {
            motoron = 0;
            ret |= 4; // ditto
        }
    }

    /* head loaded */
    if (headloaded)
        ret |= 32;

    if ((dbg>2) && (ret!=last))	// 3 polling
        fprintf(stderr,"16fdc: I34 DiskFlags %02X %s%s%s%s%s%s%s%s\n",ret,
        ret&0x80?"DRQ ":"",  ret&0x40?"":"BOOT ",ret&0x20?"HL ":"",
        ret&0x10?"":"R/O ",  ret&0x08?"M.ON ":"",ret&0x04?"M.TO ":"",
        ret&0x02?"AW.TO ":"",ret&0x01?"EOJ":"");
    last=ret;
    return ret;
}

/*  (OUT 34H)
 *	4FDC		16FDC
 * D7	AUTO WAIT	AUTO WAIT
 * D6	unassigned	DOUBLE DENSITY
 * D5	MOTOR ON	MOTOR ON
 * D4	MAXI		MAXI
 * D3	DS4		DS4
 * D2	DS3		DS3
 * D1	DS2		DS2
 * D0	DS1		DS1
 */
void cromemco_fdc_diskctl_out(BYTE data)
{
    static int last = -1;

    /* get autowait */
    autowait = (data & 128) ? 1 : 0;

    /* get selected disk */
    if      (data & 8) disk = 3;
    else if (data & 4) disk = 2;
    else if (data & 2) disk = 1;
    else               disk = 0;

    /* get disk density */
    if (data & 64) {
        ddens = DOUBLE;
        secsz = SEC_SZDD;
    } else {
        ddens = SINGLE;
        secsz = SEC_SZSD;
    }

    /* get disk format (MAXI means 8", else 5.25" */
    dtype= (data & 16)? LARGE : SMALL;

    /* motor on/off */
    if (data & 32) {
        motoron = 1;
        motortimer = 800;
        dskActv[disk]=1;
    } else {
        motoron = 0;
        motortimer = 0;
        dskActv[disk]=0;
    }
    if ((dbg>1) && (data != last))
        fprintf(stderr,"16fdc: O34 DiskCtrl %02X %s%s%s%s%c\n",data,
          data&0x80?"autowait ":"",data&0x40?"DSDD ":"",
          data&0x20?"Motor ":"",   data&0x10?"8 ":"5 ",
          (char)('A'+disk));
    last= data;
}

/*
 *	(IN 33) read data from FDC
 *	gss: if autowait, cpu into wait state until DRQ or EOJ from FDC
 *	     so after every read here (in autowait) should have DRQ or EOJ set.
 */
BYTE cromemco_fdc_data_in(void)
{
    long pos;    	/* seek position */
    int lastsec;		/* last sector of a track */

    switch (state) {
    case FDC_READ:		/* read data from disk sector */
        /* first byte? */
        if (dcnt == 0) {
            if (dbg>0)
                fprintf(stderr,"16fdc: I33 DiskRead Drv %c Tk %2d Sc %2d\n",
                    (char)('A'+disk),fdc_track,fdc_sec);
            motortimer = 800;
            dskActv[disk]=1;
            /* try to open disk image */
            dsk_path();
            strcat(fn, "/");
            strcat(fn, disks[disk].fn);
            if ((fd = open(fn, OPENRO)) == -1) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x80;	/* not ready */
                if (dbg>0) fprintf(stderr,"16fdc: I33 DiskRead FAILED - can't open %s\n",fn);
                return((BYTE) 0);
            }
            /* get drive and disk geometry */
            config_disk(fd);
            if (disks[disk].disk_t == UNKNOWN) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x10;	/* sector not found */
                close(fd);
                if (dbg>0) fprintf(stderr,"16fdc: I33 DiskRead FAILED - unknown disk type\n");
                return((BYTE) 0);
            }
            /* check track/sector */
            if ((fdc_track == 0) && (side == 0))
                lastsec = disks[disk].sec0;
            else
                lastsec = disks[disk].sectors;
            if ((fdc_track >= disks[disk].tracks) ||
                (fdc_sec == 0) ||
                (fdc_sec > lastsec)) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x10;	/* sector not found */
                close(fd);
                if (dbg>0) fprintf(stderr,
                    "16fdc: I33 DiskRead FAILED - track %d or sector %d out of range\n",
                        fdc_track,fdc_sec);
                return((BYTE) 0);
            }
            /* seek to sector */
            pos = get_pos();
            if (lseek(fd, pos, SEEK_SET) == -1L) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x10;	/* sector not found */
                close(fd);
                if (dbg>0) fprintf(stderr,"16fdc: I33 DiskRead FAILED - seek %08lX\n",pos);
                return((BYTE) 0);
            }
            /* read the sector */
            if (read(fd, &buf[0], secsz) != secsz) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x10;	/* sector not found */
                close(fd);
                if (dbg>0) fprintf(stderr,"16fdc: I33 DiskRead FAILED\n");
                return((BYTE) 0);
            }
            close(fd);
        }
        /* last byte? */
        if (dcnt == secsz - 1) {
            if (!mflag) {			/* single sector */
                state = FDC_IDLE;	/* done */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0;
                if (dbg>0) fprintf(stderr,"16fdc: I33 DiskRead completed\n");
            } else {			/* multiple sectors */
                if ((fdc_track == 0) && (side == 0))
                    lastsec = disks[disk].sec0;
                else
                    lastsec = disks[disk].sectors;
                if (lastsec == fdc_sec) {
                    state = FDC_IDLE;  /* done */
                    fdc_flags |= 1;    /* set EOJ */
                    fdc_flags &= ~128; /* reset DRQ */
                    fdc_stat = 0x10;   /* sector not found*/
                    if (dbg>0) fprintf(stderr,"16fdc: I33 DiskRead completed\n");
                } else {
                    dcnt = 0;	/* read next sector */
                    fdc_sec++;
                    return(buf[secsz - 1]);
                }
            }
        }
        if (autowait) fdc_flags |= 128; // DRQ, not yet EOJ
        /* return next byte from buffer and increment counter */
        return(buf[dcnt++]);
        break;

    case FDC_READADR:	/* read disk address */
        /* first byte? */
        if (dcnt == 0) {		/* build address field */
            if (dbg>0)
                fprintf(stderr,"16fdc: I33 DiskAddr Drv %c Tk %2d Sc %2d\n",
                    (char)('A'+disk),fdc_track,fdc_sec);
            buf[0] = fdc_track;
            buf[1] = side;
            buf[2] = fdc_sec;
            buf[3] = (ddens == SINGLE) ? 0 : 2;
            buf[4] = 0;
            buf[5] = 0;
        }
        /* last byte? */
        if (dcnt == 5) {
            state = FDC_IDLE;	/* done */
            fdc_flags |= 1;		/* set EOJ */
            fdc_flags &= ~128;	/* reset DRQ */
            fdc_stat = 0;
            if (dbg>0) fprintf(stderr,"16fdc: I33 DiskAddr complete\n");
        }
        /* return next byte from buffer and increment counter */
        return(buf[dcnt++]);
        break;

    default:
        return((BYTE) 0);
        break;
    }
}

/*
 *	(OUT 33H) write data to FDC
 */
void cromemco_fdc_data_out(BYTE data)
{
    long pos;		/* seek position */
    int lastsec;		/* last sector of a track */
    static int wrtstat;	/* state while writing (formatting) tracks */
    static int bcnt;	/* byte counter for sector data */
    static int secs;	/* # of sectors written so far */

    switch (state) {
    case FDC_WRITE:			/* write data to disk sector */
        /* first byte? */
        if (dcnt == 0) {
            if (dbg>0)
                fprintf(stderr,"16fdc: O33 DiskWrit Drv %c Tk %2d Sc %2d\n",
                    (char)('A'+disk),fdc_track,fdc_sec);
            motortimer = 800;
            dskActv[disk]=1;
            /* try to open disk image */
            dsk_path();
            strcat(fn, "/");
            strcat(fn, disks[disk].fn);
            if ((fd = open(fn, OPENRW)) == -1) {
                if ((fd = open(fn, OPENRO)) != -1) {
                    close(fd);
                    fdc_stat = 0x40; /* read only */
                } else {
                    fdc_stat = 0x80; /* not ready */
                }
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                return;
            }
            /* get drive and disk geometry */
            config_disk(fd);
            if (disks[disk].disk_t == UNKNOWN) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x10;	/* sector not found */
                close(fd);
                return;
            }
            /* check track/sector */
            if ((fdc_track == 0) && (side == 0))
                lastsec = disks[disk].sec0;
            else
                lastsec = disks[disk].sectors;
            if ((fdc_track >= disks[disk].tracks) ||
                (fdc_sec == 0) ||
                (fdc_sec > lastsec)) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x10;	/* sector not found */
                close(fd);
                return;
            }
            /* seek to sector */
            pos = get_pos();
            if (lseek(fd, pos, SEEK_SET) == -1L) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x10;	/* sector not found */
                close(fd);
                return;
            }
        }
        /* write data bytes into the sector buffer */
        buf[dcnt++] = data;
        /* last byte? */
        if (dcnt == secsz) {
            state = FDC_IDLE;		/* done */
            fdc_flags |= 1;			/* set EOJ */
            fdc_flags &= ~128;		/* reset DRQ */
            if (write(fd, &buf[0], secsz) == secsz) {
                fdc_stat = 0;
                if (dbg>0) fprintf(stderr,"16fdc: O33 DiskWrit complete\n");
            } else
                fdc_stat = 0x10;	/* sector not found */
            close(fd);
        }
        break;

    case FDC_WRTTRK:		/* write (format) track */
        if (dcnt == 0) {
            if (dbg>0)
                fprintf(stderr,"16fdc: O33 DiskWrTk Drv %c Tk %2d\n",
                    (char)('A'+disk),fdc_track);
            motortimer = 800;
            dskActv[disk]=1;
            /* unlink disk image */
            dsk_path();
            strcat(fn, "/");
            strcat(fn, disks[disk].fn);
            if ((fdc_track == 0) && (side == 0))
                unlink(fn);
            /* try to create new disk image */
            if ((fd = open(fn, OPENRW|O_CREAT, 0644)) == -1) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0x80;	/* not ready */
                return;
            }
            /* set initial drive and disk geometry */
            if ((fdc_track == 0) && (side == 0)) {
                if (dtype == LARGE) {
                    disks[disk].disk_t = LARGE;
                    disks[disk].disk_d = SINGLE;
                    disks[disk].disk_s = ONE;
                    disks[disk].tracks = TRK8;
                    disks[disk].sectors = SPT8SD;
                    disks[disk].sec0 = SPT8SD;
                } else {
                    disks[disk].disk_t = SMALL;
                    disks[disk].disk_d = SINGLE;
                    disks[disk].disk_s = ONE;
                    disks[disk].tracks = TRK5;
                    disks[disk].sectors = SPT5SD;
                    disks[disk].sec0 = SPT5SD;
                }
            }
            /* don't format tracks out of bounds */
            if (fdc_track >= disks[disk].tracks) {
                state = FDC_IDLE;    /* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0;
                close(fd);
                return;
            }
            /* now learn more */
            if (side == 1)
                disks[disk].disk_s = TWO;
            if (ddens == DOUBLE) {
                disks[disk].disk_d = DOUBLE;
                if (dtype == LARGE) {
                    disks[disk].sectors = SPT8DD;
                } else {
                    disks[disk].sectors = SPT5DD;
                }
            }
            /* seek to track */
            fdc_sec = 1;
            pos = get_pos();
            if (lseek(fd, pos, SEEK_SET) == -1L) {
                state = FDC_IDLE;	/* abort command */
                fdc_flags |= 1;		/* set EOJ */
                fdc_flags &= ~128;	/* reset DRQ */
                fdc_stat = 0;
                close(fd);
                return;
            }
            /* now wait for sector data */
            wrtstat = 1;
            secs = 0;
        }
        dcnt++;
        /* wait for sector data address mark */
        if (wrtstat == 1) {
            if (data == 0xfb) {
                wrtstat = 2;
                bcnt = 0;
            }
            return;
        }
        /* collect bytes in buffer and write if sector complete */
        if (wrtstat == 2) {
            if ((data != 0xf7) && (bcnt < secsz)) {
                buf[bcnt++] = data;
                return;
            } else {
                secs++;
                if (write(fd, buf, bcnt) == bcnt)
                    fdc_stat = 0;
                else
                    fdc_stat = 0x10; /* sector not found */
                wrtstat = 1;
            }
        }
        /* all sectors of track written? */
        if ((fdc_track == 0) && (side == 0))
            lastsec = disks[disk].sec0;
        else
            lastsec = disks[disk].sectors;
        if (secs == lastsec) {
            if (dbg>0)
                fprintf(stderr,"16fdc: O33 DiskWrTk complete\n");
            state = FDC_IDLE;
            fdc_flags |= 1;    	/* set EOJ */
            fdc_flags &= ~128;	/* reset DRQ */
            close(fd);
        }
        break;

    default:    		/* track # for seek */
        fdc_track = data;
        break;
    }
}

/*
 *	read FDC sector register
 */
BYTE cromemco_fdc_sector_in(void)
{
    return(fdc_sec);
}

/*
 *	write FDC sector register
 */
void cromemco_fdc_sector_out(BYTE data)
{
    if (dbg>1) fprintf(stderr,"16fdc: O32 DiskSector %d\n",(int)data);
    fdc_sec = data;
}

/*
 *	read FDC track register
 */
BYTE cromemco_fdc_track_in(void)
{
    return fdc_track;
}

/*
 *	write FDC track register
 *
 * Note: this does not set the FDC track register for a seek, the wanted
 * track # for a seek is written to the data port, see above.
 */
void cromemco_fdc_track_out(BYTE data)
{
    if (dbg>1) fprintf(stderr,"16fdc: O31 DiskTrack %d\n",(int)data);
//    fdc_track= data; // gss. we'll save it anyway (in case RDOS uses it)
}

/*
 *	4FDC			16FDC
 * D7	DRQ			DRQ
 * D6	SEEK IN PROGRESS	SEEK IN PROGRESS
 * D5	unassigned		unassigned
 * D4	unassigned		unassigned
 * D3	unassigned		!FP Sense Switch 5
 * D2	unassigned		!FP Sense Switch 6
 * D1	unassigned		!FP Sense Switch 7 (RDOS 2.52 boot drive)
 * D0	unassigned		!FP Sense Switch 8 (RDOS 2.52 boot drive)
 *
 * SEEK IN PROGESS is never set, we have no moving parts here.
 */
BYTE cromemco_fdc_aux_in(void) {
    static int last= -1;

    fdc_aux = 3; // xx11 = RDOS boot off drive A

    /* get DRQ from flag register */ // this is not enabled on shipped 4FDCs, ie ==0
    if (fdc_flags & 128)
        fdc_aux |= 128;
    else
        fdc_aux &= ~128;

    /* get front panel switch bits */
/*
    if ((address_switch >> 8) & 16)
        fdc_aux &= ~8;
    else
        fdc_aux |= 8;

    if ((address_switch >> 8) & 32)
        fdc_aux &= ~4;
    else
        fdc_aux |= 4;

    if ((address_switch >> 8) & 64)
        fdc_aux &= ~2;
    else
        fdc_aux |= 2;

    if ((address_switch >> 8) & 128)
        fdc_aux &= ~1;
    else
        fdc_aux |= 1;
*/
    if ((dbg>2) && (fdc_aux != last))
        fprintf(stderr,"16fdc: I04 FdcAuxIn %02X %s%s\n",fdc_aux,
          fdc_aux&0x80?"DRQ ":"",fdc_aux&0x40?"SEEKING ":"");
    last= fdc_aux;

    return fdc_aux;
}

/*
 *	4FDC		16FDC
 * D7	unassigned	unassigned
 * D6	!EJECT LEFT	!EJECT
 * D5	!EJECT RIGHT	!DRIVE SELECT OVERRIDE
 * D4	!FAST SEEK	!FAST SEEK
 * D3	!RESTORE	!RESTORE
 * D2	!CONTROL OUT	!CONTROL OUT (low pulls daisy chain down)
 * D1	unassigend	!SIDE SELECT
 * D0	unassigend	unassigned
 */
void cromemco_fdc_aux_out(BYTE data)
{
    /* fast seek for the PERSCI disk drives */
    if (!(data & 8))
        fdc_track = 0;

    /* get disk side */
    side= (data & 2) ? 0 : 1;

    if (dbg>1)
        fprintf(stderr,"16fdc: O04 FdcAuxOut %02X %s%s%s%s%s%s\n",data,
          data&0x40?"":"EJECT ", data&0x20?"":"DS.OR ",
          data&0x10?"":"FAST ",  data&0x08?"":"RESTORE ",
          data&0x04?"CO ":"!CO ",data&0x02?"H0":"H1");
}

/*
 *	read disk status
 *
 * Depends on the last command, WD status table:
 *
 * Command	D7	D6	D5	D4	D3	D2	D1	D0
 * SEEK		not	write	head	not	CRC	track	index	busy
 *		ready	protect	down	found	error	0
 *
 * STEP		not	write	head	not	CRC	track	index	busy
 * RESTORE	ready	protect	down	found	error	0
 *
 * READ		not	0	record	not	CRC	lost	DRQ	busy
 * RECORD(S)	ready		type	found	error	data
 *
 * WRITE	not	write	0	not	CRC	lost	DRQ	busy
 * RECORD(S)	read	protect		found	error	data
 *
 * READ		not	0	0	not	CRC	lost	DRQ	busy
 * ADDRESS	ready			found	error	data
 *
 * READ		not	0	0	0	0	lost	DRQ	busy
 * TRACK	ready					data
 *
 * WRITE	not	write	0	0	0	lost	DRQ	busy
 * TRACK	ready	protect				data
 */
BYTE cromemco_fdc_status_in(void)
{
    static int last = -1;

    if ((fdc_cmd&0x80) == 0) { // restore, seek, step, step in, step out
        fdc_stat &= ~0x46;
        if (disks[disk].disk_m == READONLY) fdc_stat |= 0x40;
        if (fdc_track == 0) fdc_stat |= 0x04;
        // headloaded bit was set during cromemco_fdc_cmd_out() // TODO maybe on motoroff headload=0
        if (index_pulse) { fdc_stat |= 0x02; index_pulse=0; }
    }

    if ((dbg>2) && (fdc_stat != last))	// 3=polling
        fprintf(stderr,"16fdc: I30 DiskStatus %02X %s%s%s%s%s%s%s\n",fdc_stat,
            (fdc_stat&0x80)?"Not ready ":"",
            (fdc_stat&0x40)?"R/O ":"",
            (fdc_stat&0x10)?"Not found ":"",
            (fdc_stat&0x08)?"CRC error ":"",
            (fdc_stat&0x04)?"Tk0/Lost ":"",
            (fdc_stat&0x02)?"Idx/Drq ":"",
            (fdc_stat&0x01)?"Busy ":"");


    last= fdc_stat;
    return fdc_stat;
}

//static int fdc_busy_timeout= 0;
int c16fdc_rdInt(int port) {
    if (port<DISKS) return dskActv[port];
    return 0;
}

void c16fdc_tick(void) { // every 10mS
    int i;
    static unsigned long counter = 0L;

    counter++;
    
    /* set disk index pulse */
    /* 170ms ~ 360rpm for 8" drives */
    /* 200ms ~ 300rpm for 5.25" drives */
    if ((counter % (dtype == LARGE)? 17 : 20) == 0) index_pulse = 1;

    /* motor timeout timer */
    if (motortimer > 0) {
        if (--motortimer == 0) for (i=0; i<DISKS; i++) dskActv[i]=0;
    }

/*
    if (fdc_busy_timeout)
        if (--fdc_busy_timeout == 0)  {
            fdc_stat &= ~1;
            // and maybe set EOJ and timedout
        }
*/
}

/*
 *	write disk command
 */
void cromemco_fdc_cmd_out(BYTE data)
{
    /* CDOS 1.07 issues a GetAddr then issues another command.
       Whilst the chip should refuse the command, as it is BUSY, waiting for
       the requested bytes to be transferred; it seems that it just moves on to
       the next command. This might just be with the older WD1771 chip / 4FDC
       card. - gss, 26Nov17 */
#if 0
    /* if controller busy ignore all commands but interrupt */
    if ((fdc_stat & 1) && ((data & 0xf0) != 0xd0)) {
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X ignored - busy\n",data);
        return;
    }
#endif

    fdc_cmd = data;     // save last cmd. status varies with this.

    /* new command, reset EOJ */
    fdc_flags &= ~1;

    if ((data & 0xf0) == 0) {		/* restore (seek track 0) */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X restore\n",data);
        headloaded = (data & 8) ? 1 : 0;
        fdc_track = 0;
        fdc_flags |= 1;			/* set EOJ */
        fdc_stat = 4;			/* positioned to track 0 */
        if (headloaded) fdc_stat |= 32;

    } else if ((data & 0xf0) == 0x10) {	/* seek */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X seek\n",data);
        headloaded = (data & 8) ? 1 : 0;
        fdc_flags |= 1;    		/* set EOJ */
        if (fdc_track <= disks[disk].tracks)
            fdc_stat = (fdc_track == 0) ? 4 : 0;
        else
            fdc_stat = 16;		/* seek error */
        if (headloaded) fdc_stat |= 32;

    } else if ((data & 0xe0) == 0x20) {	/* step */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X step\n",data);
        headloaded = (data & 8) ? 1 : 0;
        //if (data & 0x10)
            fdc_track += step_dir;
        fdc_flags |= 1;    		/* set EOJ */
        if (fdc_track <= disks[disk].tracks)
            fdc_stat = (fdc_track == 0) ? 4 : 0;
        else
            fdc_stat = 16;		/* seek error */
        if (headloaded) fdc_stat |= 32;

    } else if ((data & 0xe0) == 0x40) {	/* step in */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X step in\n",data);
        headloaded = (data & 8) ? 1 : 0;
        //if (data & 0x10)
            fdc_track++;
        fdc_flags |= 1;    		/* set EOJ */
        step_dir = 1;
        if (fdc_track <= disks[disk].tracks)
            fdc_stat = (fdc_track == 0) ? 4 : 0;
        else
            fdc_stat = 16;		/* seek error */
        if (headloaded) fdc_stat |= 32;

    } else if ((data & 0xe0) == 0x60) {	/* step out */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X step out\n",data);
        headloaded = (data & 8) ? 1 : 0;
        //if (data & 0x10)
            fdc_track--;
        fdc_flags |= 1;			/* set EOJ */
        step_dir = -1;
        if (fdc_track <= disks[disk].tracks)
            fdc_stat = (fdc_track == 0) ? 4 : 0;
        else
            fdc_stat = 16;    	/* seek error */
        if (headloaded) fdc_stat |= 32;

    } else if ((data & 0xf0) == 0xd0) {	/* force interrupt */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X interrupt\n",data);
        state = FDC_IDLE;		/* abort any command */
        fdc_stat = 0;
        fdc_flags &= ~128;    	/* clear DRQ */
        if (data & 0x0f) fdc_flags |= 1;		/* set EOJ */

    } else if ((data & 0xe0) == 0x80) {	/* read sector(s) */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X read sector\n",data);
        mflag = (data & 16) ? 1 : 0;
        state = FDC_READ;
        dcnt = 0;
        fdc_stat = 3;			/* set DRQ & busy */
        fdc_flags |= 128;		/* set DRQ */

    } else if ((data & 0xf0) == 0xa0) {	/* write single sector */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X write sector\n",data);
        state = FDC_WRITE;
        dcnt = 0;
        fdc_stat = 3;			/* set DRQ & busy */
        fdc_flags |= 128;		/* set DRQ */

    } else if (data == 0xc4) {		/* read address */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X read address\n",data);
        state = FDC_READADR;
        dcnt = 0;
        fdc_stat = 3;			/* set DRQ & busy */
        //fdc_flags |= 128;		/* set DRQ */
        fdc_flags |= 129; /* RDOS 2&3 seem to need EOJ already, why */

    } else if ((data & 0xf0) == 0xe0) {	/* read track */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X read track - not implemented\n",data);
        fdc_stat = 16;			/* not found */
        fdc_flags |= 1;			/* set EOJ */

    } else if ((data &0xf0) == 0xf0) {	/* write track */
        if (dbg>0) fprintf(stderr,"16fdc: O30 DiskCmd %02X write track\n",data);
        state = FDC_WRTTRK;
        dcnt = 0;
        fdc_stat = 3;			/* set DRQ & busy */
        fdc_flags |= 128;		/* set DRQ */

    } else {
        fprintf(stderr,"16fdc: O30 DiskCmd %02X unknown command\n", data);
        fdc_stat = 16|8;		/* not found & CRC error */
        fdc_flags |= 1;			/* set EOJ */
    }
}

/*
 * Reset FDC, placeholder for now, consult manuals
 */
void cromemco_fdc_reset(void)
{
}
